//***************************************************************************
// Ces applets viennent en complment du livre :
//
//  		VIE ARTIFICIELLE
//      O la biologie rencontre l'informatique
//		Illustr en Java
//            Jean-Philippe Rennard
//           	  Vuibert 2002
//===========================================================================
//  Algorithme Gntique : Problme du Voyageur de Commerce  (Vuibert)
//    v:0.1 10/09/2002
//
//    This program is free software; you can redistribute it and/or
//    modify it under the terms of the GNU General Public License
//    as published by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//***************************************************************************
package pvcag;

import java.awt.*;
import java.util.*;
import vabase.*;

/** classe de gestion de l'algorithme gntique */
public class Cag extends Cprocess {
  /** paramtres par dfaut */
  static final int NB_VILLES=20;
  static final int POP=100;
  static final float MUTATION=0.05F;
  /** Mode d'chalonnage */
  static final int SCALENONE=0; // aucun
  static final int SCALELINEAR=1; // linaire
  static final int SCALECLASSE=2; // Ranking
  static final int SCALESIGMA=3;  // Sigma truncation
  /** Mode crossover */
  static final int INVERSION=0;
  static final int CPA=1;
  /** Pointeur sur l'interface utilisateur */
  private pvcUI appUI;
  /** Vue */
  vuePvc pvcVue;
  /** Affichage pire parcours */
  private  boolean modeAffPire=false;
  /** Tableau des villes */
  Cville villes[];
  /** Population */
  private Vector lesIndividus;

  /** paramtres */
  int nbVilles;
  int taillePopulation;
  float tauxMutation=Cag.MUTATION;
  /** Mode litiste */
  private boolean modeElitiste=true;
  /** Mode d'chalonnage */
  private int modeEchel=SCALELINEAR;
  /** Mode de crossover */
  private int modeCross=INVERSION;
  /** Meilleure fitness */
  private double meilleurFitness;
  /** Meilleur parcours */
  private double meilleurFitnessBrute;
  /** Pointeur sur meilleur individu */
  private Cindividu meilleurInd;
  /** Fitness la plus mauvaise */
  private double pireFitness;
  /** Parcours le plus long */
  private double pireFitnessBrute;
  /** Pointeur sur pire individu */
  private Cindividu pireInd;
  /** Fitness moyenne */
  private double moyenneFitness;
  /** Ecart-type de la Fitness */
  private double ecartFitness;
  /** Somme des fitness */
  private double sommeFitness;

  /** Construction de l'ag dans une interface utilisateur, selon nombre de
   *  villes, population et taux de mutation */
  public Cag(pvcUI ui, int nbv, int pop, float mut) {
    super(ui);
    appUI=ui;
    pvcVue=new vuePvc(this, Color.black, Color.blue);
    vue=(Cvue)pvcVue;
    setNewPb(nbv,pop, mut);
  }
  /** Initialisation d'un nouveau problme */
  public void setNewPb(int nbv, int pop, float mut) {
    nbVilles=nbv;
    tauxMutation=mut;
    creeVilles(nbVilles);
    setNewPop(pop);
  }
  /** Cration d'une nouvelle population */
  public void setNewPop(int p) {
    taillePopulation=p;
    generation=0;
    creePop();
    setGenData();
  }
  /** Initialisation/MAJ de la vue */
  public void initVue() {
    pvcVue.setNewTaille();
  }
  /** MAJ mode */
  public void setMode(boolean m) {
    modeElitiste=m;
  }
  /** MAJ type crossover */
  public void setCross(int cross) {
    modeCross=cross;
  }
  /** Choix affichage pire parcours */
  public void setAffPire(boolean m) {
    modeAffPire=m;
  }
  /** MAJ echelonnage */
  public void setEchel(int m) {
    modeEchel=m;
  }
  /** Affichage parcours */
  public void affParcours() {
    pvcVue.affParcours(meilleurInd.getGenome(), Color.black);
    if(modeAffPire)
      pvcVue.affParcours(pireInd.getGenome(), Color.red);
  }

  /** Cration d'une nouvelle gnration */
  public void geneSuiv() {
    creeNewPop();
    setGenData();
    generation++;
    pvcVue.peintFond();
    affParcours();
    pvcVue.repaint();
    appUI.affDonnees(ecartFitness,meilleurFitnessBrute, pireFitnessBrute);
  }

  /** Cration de la population initiale */
  private void creePop() {
    lesIndividus=null;
    lesIndividus=new Vector(taillePopulation);
    for(int i=0;i<taillePopulation;i++)
      lesIndividus.addElement(new Cindividu(nbVilles));
  }
  /** Cration d'une nouvelle population */
  private void creeNewPop() {
    // Vecteur tampon
    Vector popTmp=new Vector(taillePopulation);
    // mode elitiste
    int rang=0;
    if(modeElitiste) {
      popTmp.addElement(meilleurInd);
      rang++;
    }
    for(int i=rang;i<taillePopulation;i++) {
      Cindividu ind=new Cindividu(selectIndividu(), selectIndividu(), modeCross);
      popTmp.addElement(ind);
      ind.mute(tauxMutation);
    }
    lesIndividus=popTmp;
  }
  /** Slection d'un individu selon roue de loterie */
  private Cindividu selectIndividu() {
    double val=Math.random()*sommeFitness;
    double sommePartielle=0;
    int i=0;
    do {
      sommePartielle += ((Cindividu)lesIndividus.elementAt(i++)).getFitness();
    } while(sommePartielle<val && i<=taillePopulation);
    return (Cindividu)lesIndividus.elementAt(--i);
  }
  /** Mise  jour des donnes de la gnration */
  private void setGenData() {
    calcFitnessBrutePop();
    calcPopStats();
    echelonPop();
  }
  /** Calcul et MAJ fitness brute d'un individu */
  private void calcFitnessBrute(Cindividu ind) {
    long fit=0;
    for(int i=1;i<nbVilles+1;i++)
      fit += (long) Math.sqrt((Math.pow(villes[ind.getGenome().charAt(i-1)].x-
                      villes[ind.getGenome().charAt(i)].x,2)+
                      Math.pow(villes[ind.getGenome().charAt(i-1)].y-
                      villes[ind.getGenome().charAt(i)].y,2)));
    ind.setFitnessBrute(fit);
  }
  /** Cration des villes */
  private void creeVilles(int nbv) {
    villes=null;
    villes=new Cville[nbv];
    for(int i=0;i<nbv;i++)
      villes[i]=new Cville(pvcVue.lLargeur, pvcVue.lHauteur);
  }
  /** Calcul Fitness brute de la population */
  private void calcFitnessBrutePop() {
    long maxFitBr = Long.MIN_VALUE;
    long minFitBr = Long.MAX_VALUE;
    Cindividu ind;
    // Pour chaque individu...
    for(int i=0;i<taillePopulation;i++) {
      ind=(Cindividu)lesIndividus.elementAt(i);
      // calcule sa fitness brute
      calcFitnessBrute(ind);
      if(ind.getFitnessBrute()>maxFitBr)
        maxFitBr=ind.getFitnessBrute();
      if(ind.getFitnessBrute()<minFitBr)
        minFitBr=ind.getFitnessBrute();
    }
    meilleurFitnessBrute=minFitBr;
    pireFitnessBrute=maxFitBr;
    // chelonne Fitness par rapport au maxi de la gnration
    for(int i=0;i<taillePopulation;i++) {
      ind=(Cindividu)lesIndividus.elementAt(i);
      ind.setFitness(maxFitBr-ind.getFitnessBrute());
    }
  }
  /** chelonnement de la population */
  private void echelonPop() {
    Cindividu ind;
    switch(modeEchel) {
      // linaire
      case SCALELINEAR:
        double fmul=2.0;
        double delta=0.0;
        double transLinA, transLinB;
        if(pireFitness>(fmul*moyenneFitness-meilleurFitness)/(fmul-1.0)) {
          delta=meilleurFitness-moyenneFitness;
          transLinA=(fmul-1.0)*moyenneFitness/delta;
          transLinB=moyenneFitness*(meilleurFitness-fmul*moyenneFitness)/delta;
        }
        else {
          delta=moyenneFitness-pireFitness;
          transLinA=moyenneFitness/delta;
          transLinB=-1.0*pireFitness*moyenneFitness/delta;
        }
        for(int i=0;i<taillePopulation;i++) {
          ind=(Cindividu)lesIndividus.elementAt(i);
          ind.setFitness((double)(transLinA*ind.getFitness()+transLinB));
        }
        calcPopStats();
        break;
      // Ranking
      case SCALECLASSE:
        Sort(lesIndividus, -1);
        for(int i=0;i<taillePopulation;i++) {
          ind=(Cindividu)lesIndividus.elementAt(i);
          ind.setFitness((double)Math.max(taillePopulation-i,1));
        }
        calcPopStats();
        break;
      // Sigma truncation
      case SCALESIGMA:
        double s;
        for(int i=0;i<taillePopulation;i++) {
          ind=(Cindividu)lesIndividus.elementAt(i);
          if(ecartFitness!=0)
            s=1+((ind.getFitness()-moyenneFitness)/(2*ecartFitness));
          else
            s=1;
          if(s<0)
            s=0.1;
          ind.setFitness(s);
        }
        calcPopStats();
        break;
      // Aucun
      case SCALENONE:
      default:
        break;
    }
  }

  /** Calcul des statistiques de la population */
  private void calcPopStats() {
    double meilFit= Long.MIN_VALUE;
    double pireFit= Long.MAX_VALUE;
    double sommeFit=0;
    double sommeFitCarre=0;
    Cindividu ind;
    // calculs stats
    for(int i=0;i<taillePopulation;i++) {
      ind=(Cindividu)lesIndividus.elementAt(i);
      // dtermination des valeurs extrmes
      if(ind.getFitness()>meilFit) {
        meilFit=ind.getFitness();
        meilleurInd=ind;
      }
      if(ind.getFitness()<pireFit) {
        pireFit=ind.getFitness();
        pireInd=ind;
      }
      // sommes intermdiaires
      sommeFit += ind.getFitness();
      sommeFitCarre += Math.pow(ind.getFitness(),2);
    }
    // fixation des extrmes
    meilleurFitness=meilFit;
    pireFitness=pireFit;
    sommeFitness=sommeFit;
    // calcul moyenne et cart-type
    moyenneFitness=(double)sommeFit/(double)taillePopulation;
    ecartFitness=Math.sqrt(
                      (sommeFitCarre/taillePopulation)-Math.pow(moyenneFitness,2));
  }

  /** Appel tri rapide. sens=1 : ascendant, -1 : descendant */
  private void Sort(Vector v, int sens) {
    v.trimToSize();
    QuickSort(v,0,v.size()-1,sens);
  }
  /** Swap element i and j in Vector v */
  private void Swap(Vector v,int i, int j) {
    Object o=v.elementAt(i);
    v.setElementAt(v.elementAt(j),i);
    v.setElementAt(o,j);
  }
  /** Quick sort algorithm */
  private void QuickSort(Vector v,int low, int up, int sens) {
    int lo = low;
    int hi = up;
    double mid;

    if ( up > low) {
      mid = ((Cindividu)(v.elementAt((low+up)/2))).getFitness()*sens;
      while( lo <= hi ) {
	      while( ( lo < up ) && ( ((Cindividu)(v.elementAt(lo))).getFitness()*sens < mid ))
          ++lo;
	      while( ( hi > low ) && ( ((Cindividu)(v.elementAt(hi))).getFitness()*sens > mid ))
    		  --hi;
        if( lo <= hi ) {
          Swap(v, lo, hi);
          ++lo;
          --hi;
        }
      }
      if( low < hi )
        QuickSort( v, low, hi, sens );
      if( lo < up )
        QuickSort( v, lo, up, sens );
    }
  }
}